/* GIMP Plug-in Toy
 * Copyright (C) 2011  Rüdiger Schneider <remschneid@web.de> (the "Author").
 * All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the Author of the
 * Software shall not be used in advertising or otherwise to promote the
 * sale, use or other dealings in this Software without prior written
 * authorization from the Author.
 */

#include "config.h"
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <gdk/gdkkeysyms.h>
#include "main.h"
#include "interface.h"
#include "plugin-intl.h"

gboolean toy_dialog(gint32 image_ID,GimpDrawable *drawable)
{
	GtkWidget *dialog;
	GtkWidget *hbox1;
	GtkWidget *hbox2;
	GtkWidget *vbox;
	GtkWidget *table;
	GtkWidget *frame1;
	GtkWidget *frame2;
	GtkWidget *preview;
	GtkWidget *gradientlabel;
	GtkObject *radius;
	GtkObject *contrast;
	GtkObject *saturation;
	GtkObject *startx;
	GtkObject *starty;
	GtkObject *endx;
	GtkObject *endy;	
	gboolean run,isrgb;
	
	tmpvals.drawable=drawable;
	toy_variables_init();
	isrgb=gimp_drawable_is_rgb(drawable->drawable_id);
	gimp_ui_init(PROCEDURE_NAME, FALSE);
	dialog=gimp_dialog_new(_("Toy"),PROCEDURE_NAME,NULL,0,gimp_standard_help_func,PROCEDURE_NAME,GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,GTK_STOCK_OK, GTK_RESPONSE_OK,NULL);	
	g_signal_connect(dialog,"key_press_event",G_CALLBACK(pressed_keys),NULL);
	g_signal_connect(dialog,"key_release_event",G_CALLBACK(released_keys),NULL); 
	gimp_window_set_transient(GTK_WINDOW (dialog));
	hbox1=gtk_hbox_new(FALSE, 12);
	gtk_container_set_border_width(GTK_CONTAINER (hbox1), 12);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),hbox1);
	gtk_widget_show(hbox1);	
	vbox=gtk_vbox_new(FALSE, 12);
	gtk_container_add(GTK_CONTAINER(hbox1),vbox);
	gtk_widget_show(vbox);
	frame1=gtk_aspect_frame_new(NULL,0.0,0.0,1.0,TRUE);
	gtk_frame_set_shadow_type(GTK_FRAME(frame1),GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(vbox),frame1,TRUE,TRUE,0);
	gtk_widget_show(frame1);
	tmpvals.preview=preview=gimp_preview_area_new();
	gtk_widget_set_size_request(preview,tmpvals.preview_width,tmpvals.preview_height);
	gtk_container_add(GTK_CONTAINER(frame1),preview);
	gtk_widget_show(preview);
	gtk_widget_add_events(preview,GDK_POINTER_MOTION_MASK|GDK_BUTTON1_MOTION_MASK|GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK|GDK_POINTER_MOTION_HINT_MASK);
	g_signal_connect(preview,"realize",G_CALLBACK(toy_realize_callback),NULL);	
	g_signal_connect(preview, "motion_notify_event",G_CALLBACK(motion_and_mouse),NULL);
	g_signal_connect(preview, "button_press_event",G_CALLBACK(motion_and_mouse),NULL);
	g_signal_connect(preview, "button_release_event",G_CALLBACK(mouse_released),NULL);
	g_signal_connect_after(preview, "expose-event",G_CALLBACK (toy_expose_callback),NULL);  	
  	gradientlabel=gimp_hint_box_new(_("Click and drag the knots in the preview to define the depth of field. Shift click and drag the knots to position the focus. Press the space bar to toggle visibility of the gradient handles."));
	gtk_box_pack_end (GTK_BOX(vbox),gradientlabel, FALSE, FALSE, 0);
	gtk_widget_show(gradientlabel);
	hbox2=gtk_hbox_new(FALSE,12);
	gtk_box_pack_start(GTK_BOX (hbox1),hbox2,FALSE,FALSE,0);
	gtk_widget_show(hbox2);
	frame2=gimp_frame_new(_("Picture parameters"));
	gtk_box_pack_start(GTK_BOX (hbox2),frame2,FALSE,FALSE,0);
	gtk_widget_show(frame2);
	table=gtk_table_new((isrgb)?7:6,3,FALSE);
	gtk_table_set_col_spacings(GTK_TABLE(table),6);
	gtk_table_set_row_spacings(GTK_TABLE(table),6);
	gtk_container_add(GTK_CONTAINER(frame2),table);
	gtk_widget_show(table);
	radius=gimp_scale_entry_new(GTK_TABLE(table),0,1,_("Blur:"),200,5,toyvals.radius,5.0,100.0,0.1,1.0,1,TRUE,0,0,NULL,NULL);
	g_signal_connect(radius, "value-changed",G_CALLBACK(gimp_double_adjustment_update_r),&toyvals.radius);
	contrast=gimp_scale_entry_new(GTK_TABLE(table),0,2,_("Contrast:"),200,5,toyvals.contrast, 0,127,1,10,0,TRUE,0,0,NULL,NULL);
	g_signal_connect(contrast, "value-changed",G_CALLBACK (gimp_int_adjustment_update_cs),&toyvals.contrast);	
	if (isrgb)
		{
		saturation=gimp_scale_entry_new(GTK_TABLE(table),0,3,_("Saturation:"),200,5,toyvals.saturation, 0,100,1,10,0,TRUE,0,0,NULL,NULL);
		g_signal_connect(saturation, "value-changed",G_CALLBACK(gimp_int_adjustment_update_cs),&toyvals.saturation);
		}
	startx=gimp_scale_entry_new(GTK_TABLE(table),0,4,_("Start X:"),200,5,toyvals.start[0],0.0,1.0,0.01,0.1,2,TRUE,0,0,NULL,NULL);
	knotvals.startx=GTK_ADJUSTMENT(startx);
	g_signal_connect(startx, "value-changed",G_CALLBACK(gimp_double_adjustment_update_m),&toyvals.start[0]);
	starty=gimp_scale_entry_new(GTK_TABLE(table),0,5,_("Start Y:"),200,5,toyvals.start[1],0.0,1.0,0.01,0.1,2,TRUE,0,0,NULL,NULL);
	knotvals.starty=GTK_ADJUSTMENT(starty);
	g_signal_connect(starty, "value-changed",G_CALLBACK(gimp_double_adjustment_update_m),&toyvals.start[1]);
	endx=gimp_scale_entry_new(GTK_TABLE(table),0,6,_("End X:"),200,5,toyvals.end[0],0.0,1.0,0.01,0.1,2,TRUE,0,0,NULL,NULL);
	knotvals.endx=GTK_ADJUSTMENT(endx);
	g_signal_connect(endx, "value-changed",G_CALLBACK(gimp_double_adjustment_update_m),&toyvals.end[0]);
	endy=gimp_scale_entry_new(GTK_TABLE(table),0,7,_("End Y:"),200,5,toyvals.end[1],0.0,1.0,0.01,0.1,2,TRUE,0,0,NULL,NULL);
	knotvals.endy=GTK_ADJUSTMENT(endy);
	g_signal_connect(endy, "value-changed",G_CALLBACK(gimp_double_adjustment_update_m),&toyvals.end[1]);
	gtk_widget_show (dialog);
	toy_value_changed(preview);
	run=(gimp_dialog_run(GIMP_DIALOG (dialog))==GTK_RESPONSE_OK);
	gtk_widget_destroy (dialog);
	free_buffers_and_tempimage();
	return run;
}

void draw_gradient(GimpDrawable *drawable,GimpPreviewArea *preview)
{
	cairo_t *cr;
	if (knotvals.gradient_hidden) {return;}
	gint mx=(gint)(ROUND(toyvals.start[0]*tmpvals.preview_width));
	gint my=(gint)(ROUND(toyvals.start[1]*tmpvals.preview_height));
	gint ex=(gint)(ROUND(toyvals.end[0]*tmpvals.preview_width));
	gint ey=(gint)(ROUND(toyvals.end[1]*tmpvals.preview_height));
	gint sx=mx+(mx-ex);
	gint sy=my+(my-ey);
	cr=gdk_cairo_create(GTK_WIDGET(preview)->window);	
	cairo_set_source_rgb (cr,1,1,1);
	cairo_set_operator(cr,CAIRO_OPERATOR_ADD);//CAIRO_OPERATOR_DIFFERENCE);
	cairo_move_to(cr,sx,sy);
	cairo_line_to(cr,ex,ey);
	cairo_stroke(cr);
	//cairo_set_operator(cr,CAIRO_OPERATOR_ADD);
	cairo_set_source_surface(cr,knotvals.endknot,sx-(knotvals.knotsize/2),sy-(knotvals.knotsize/2));
	cairo_mask_surface(cr,knotvals.endknot,sx-(knotvals.knotsize/2),sy-(knotvals.knotsize/2));
	cairo_set_operator(cr,CAIRO_OPERATOR_SOURCE);
	cairo_set_source_surface(cr,knotvals.middleknot,mx-(knotvals.knotsize/2),my-(knotvals.knotsize/2));
	cairo_mask_surface(cr,knotvals.middleknot,mx-(knotvals.knotsize/2),my-(knotvals.knotsize/2));
	cairo_set_source_surface(cr,knotvals.startknot,ex-(knotvals.knotsize/2),ey-(knotvals.knotsize/2));
	cairo_mask_surface(cr,knotvals.startknot,ex-(knotvals.knotsize/2),ey-(knotvals.knotsize/2));
	cairo_destroy(cr);
}

void toy_variables_init()
{
	cairo_t *cr;
	gchar token[30];
	gchar *value=NULL,*end=NULL;	
	gdouble factor_x,factor_y;
	gint16 maxpreview=MAXPREVIEWSIZE;
	
	sprintf(token,"%s_previewsize",PROCEDURE_NAME);
	value=gimp_gimprc_query(token);
	if (value)
		{
		gdouble val=g_ascii_strtod(value,&end);
		maxpreview=(gint16)val;
		if ((!maxpreview)||(maxpreview>2048)) {maxpreview=MAXPREVIEWSIZE;}
		g_free (value);
		
		}
	sprintf(token,"%s_knotsize",PROCEDURE_NAME);
	value=gimp_gimprc_query(token);
	if (value)
		{
		gint8 tmpvalue;	
		gdouble val=g_ascii_strtod(value,&end);
		tmpvalue=(gint8)val;
		if ((tmpvalue>=5)||(tmpvalue<=255)) {knotvals.knotsize=tmpvalue;}
		g_free (value);
		}
	factor_x=(gdouble)tmpvals.drawable->width/maxpreview;
	factor_y=(gdouble)tmpvals.drawable->height/maxpreview;
	tmpvals.pre2img=(factor_x>=factor_y)?factor_x:factor_y;
	if ((factor_x<=1.0)&&(factor_y<=1.0)) {tmpvals.pre2img=1.0;}
	tmpvals.preview_width=(gint)(tmpvals.drawable->width/tmpvals.pre2img);
	tmpvals.preview_height=(gint)(tmpvals.drawable->height/tmpvals.pre2img);
	knotvals.coordinates[0]=(gint)(ROUND(toyvals.start[0]*tmpvals.preview_width));
	knotvals.coordinates[1]=(gint)(ROUND(toyvals.start[1]*tmpvals.preview_height));
	knotvals.coordinates[2]=(gint)(ROUND(toyvals.end[0]*tmpvals.preview_width));
	knotvals.coordinates[3]=(gint)(ROUND(toyvals.end[1]*tmpvals.preview_height));
	knotvals.difference_x=knotvals.coordinates[2]-knotvals.coordinates[0];
	knotvals.difference_y=knotvals.coordinates[3]-knotvals.coordinates[1];
	if (!knotvals.middleknot)
		{
		knotvals.middleknot=cairo_image_surface_create(CAIRO_FORMAT_RGB24,knotvals.knotsize,knotvals.knotsize);
		cr=cairo_create(knotvals.middleknot);
		cairo_set_source_rgb (cr,1,1,1);
		cairo_rectangle(cr,0,0,knotvals.knotsize,knotvals.knotsize);
		cairo_fill(cr);
		cairo_set_source_rgb (cr,0.2,0.8,0.2);
		cairo_rectangle(cr,1,1,knotvals.knotsize-2,knotvals.knotsize-2);
		cairo_fill(cr);
		cairo_destroy(cr);
		}	
	if (!knotvals.startknot)
		{
		knotvals.startknot=cairo_image_surface_create(CAIRO_FORMAT_RGB24,knotvals.knotsize,knotvals.knotsize);
		cr=cairo_create(knotvals.startknot);
		cairo_set_source_rgb (cr,1,1,1);
		cairo_rectangle(cr,0,0,knotvals.knotsize,knotvals.knotsize);
		cairo_fill(cr);
		cairo_set_source_rgb (cr,0.8,0.2,0.2);
		cairo_rectangle(cr,1,1,knotvals.knotsize-2,knotvals.knotsize-2);
		cairo_fill(cr);
		cairo_destroy(cr);
		}
	if (!knotvals.endknot)
		{
		knotvals.endknot=cairo_image_surface_create(CAIRO_FORMAT_RGB24,knotvals.knotsize,knotvals.knotsize);
		cr=cairo_create(knotvals.endknot);
		cairo_set_source_rgb (cr,1,1,1);
		cairo_move_to(cr,knotvals.knotsize/2,0);
		cairo_line_to(cr,knotvals.knotsize/2,knotvals.knotsize-1);
		cairo_stroke(cr);
		cairo_move_to(cr,0,knotvals.knotsize/2);
		cairo_line_to(cr,knotvals.knotsize-1,knotvals.knotsize/2);
		cairo_stroke(cr);
		cairo_destroy(cr);
		}
	// To ensure that preview is updated at dialog appearance
	preview_idle=1;	
}

void toy_realize_callback(GtkWidget *widget)
{
	GdkDisplay *display=gtk_widget_get_display(widget);
	GdkCursor *cursor=gdk_cursor_new_for_display(display,GDK_CROSSHAIR);
	gdk_window_set_cursor(widget->window,cursor);
	gdk_cursor_unref(cursor);
}

gboolean toy_expose_callback(GtkWidget *widget,GdkEvent *event)
{
	draw_gradient(tmpvals.drawable,GIMP_PREVIEW_AREA(tmpvals.preview));
	return FALSE;
}

void mouse_released(GtkWidget *widget,GdkEvent *event)
{
	if (knotvals.button_down)
		{
		knotvals.coordinates[0]=CLAMP(knotvals.coordinates[0],0,tmpvals.preview_width-1);
		knotvals.coordinates[1]=CLAMP(knotvals.coordinates[1],0,tmpvals.preview_height-1);
		knotvals.coordinates[2]=CLAMP(knotvals.coordinates[2],0,tmpvals.preview_width-1);
		knotvals.coordinates[3]=CLAMP(knotvals.coordinates[3],0,tmpvals.preview_height-1);
		knotvals.difference_x=knotvals.coordinates[2]-knotvals.coordinates[0];
		knotvals.difference_y=knotvals.coordinates[3]-knotvals.coordinates[1];
		knotvals.button_down=FALSE;
		}
}

gboolean motion_and_mouse(GtkWidget *widget,GdkEvent *event)
{
	GdkModifierType state;
	gint x,y,size=knotvals.knotsize/2;
	gboolean shifted,pressed;
	
	if (knotvals.gradient_hidden) 
		{
		knotvals.difference_x=0;
		knotvals.difference_y=0;
		knotvals.inknot1=FALSE;
		knotvals.inknot2=FALSE;
		return TRUE;
		}

	gdk_window_get_pointer(widget->window,&x,&y,&state);
	shifted=(state&GDK_SHIFT_MASK);
	pressed=(state&GDK_BUTTON1_MASK);

	if (knotvals.button_down&&pressed) {recalc_knot_position(x,y,shifted);}
	
	if ((!shifted&&knotvals.button_down)||(!knotvals.button_down))
		{
		knotvals.inknot1=((x>=knotvals.coordinates[0]-size)&&(x<=knotvals.coordinates[0]+size)&&(y>=knotvals.coordinates[1]-size)&&(y<=knotvals.coordinates[1]+size));
		knotvals.inknot2=((x>=knotvals.coordinates[2]-size)&&(x<=knotvals.coordinates[2]+size)&&(y>=knotvals.coordinates[3]-size)&&(y<=knotvals.coordinates[3]+size));	
		}
	
	GdkDisplay *display=gtk_widget_get_display(widget);
	GdkCursor *cursor=gdk_cursor_new_for_display(display,((!knotvals.inknot1)&&(!knotvals.inknot2))?GDK_CROSSHAIR:(shifted)?GDK_FLEUR:GDK_HAND1);
	gdk_window_set_cursor(widget->window,cursor);
	gdk_cursor_unref(cursor);
	
	if (pressed&&!knotvals.button_down)
		{
		gint mx=(gint)(ROUND(toyvals.start[0]*tmpvals.preview_width));
		gint my=(gint)(ROUND(toyvals.start[1]*tmpvals.preview_height));
		gint ex=(gint)(ROUND(toyvals.end[0]*tmpvals.preview_width));
		gint ey=(gint)(ROUND(toyvals.end[1]*tmpvals.preview_height));
		if (knotvals.inknot1)
			{
			knotvals.offsetx=x-mx;
			knotvals.offsety=y-my;
			knotvals.button_down=TRUE;
			}
		else if (knotvals.inknot2)
			{
			knotvals.offsetx=x-ex;
			knotvals.offsety=y-ey;
			knotvals.button_down=TRUE;
			}
		}
		
return FALSE;
}

void recalc_knot_position(gint x,gint y,gboolean both)
{
	gint w=tmpvals.preview_width;
	gint h=tmpvals.preview_height;
	gint active_knot_x=x+knotvals.offsetx;
	gint active_knot_y=y+knotvals.offsety;	
	gint other_knot_x=0,other_knot_y=0;
	
	if (both&&knotvals.inknot1)
		{
		other_knot_x=active_knot_x+knotvals.difference_x;
		other_knot_y=active_knot_y+knotvals.difference_y;
		}
	else if(both&&knotvals.inknot2)
		{
		other_knot_x=active_knot_x-knotvals.difference_x;
		other_knot_y=active_knot_y-knotvals.difference_y;
		}
		
	if (knotvals.inknot1)
		{
		if (both)
			{	
			if (knotvals.difference_x>=0)
				{
				if (active_knot_x>=w-knotvals.difference_x-1) 
					{
					active_knot_x=w-knotvals.difference_x-1;
					other_knot_x=w-1;
					}
				else if (active_knot_x<=0)
					{
					active_knot_x=0;
					other_knot_x=knotvals.difference_x;
					}
				}
			else	//knotvals.difference_x<0
				{
				if (active_knot_x>=w-1) 
					{					
					active_knot_x=w-1;
					other_knot_x=w+knotvals.difference_x-1;
					}
				else if (active_knot_x<=-knotvals.difference_x)
					{
					active_knot_x=-knotvals.difference_x;
					other_knot_x=0;
					}
				}
			if (knotvals.difference_y>=0)
				{
				if (active_knot_y>=h-knotvals.difference_y-1) 
					{
					active_knot_y=h-knotvals.difference_y-1;
					other_knot_y=h-1;
					}
				else if (active_knot_y<=0)
					{
					active_knot_y=0;
					other_knot_y=knotvals.difference_y;	
					}
				}
			else //knotvals.difference_y<0
				{
				if (active_knot_y>=h-1) 
					{					
					active_knot_y=h-1;
					other_knot_y=h+knotvals.difference_y-1;
					}
				else if (active_knot_y<=-knotvals.difference_y)
					{
					active_knot_y=-knotvals.difference_y;
					other_knot_y=0;
					}
				}
			toyvals.end[0]=CLAMP((gdouble)other_knot_x/(gdouble)(tmpvals.preview_width),0.0,1.0);
			toyvals.end[1]=CLAMP((gdouble)other_knot_y/(gdouble)(tmpvals.preview_height),0.0,1.0);
			gtk_adjustment_set_value(knotvals.endx,toyvals.end[0]);
			gtk_adjustment_set_value(knotvals.endy,toyvals.end[1]);
			knotvals.coordinates[2]=other_knot_x;
			knotvals.coordinates[3]=other_knot_y;
			}
		toyvals.start[0]=CLAMP((gdouble)active_knot_x/(gdouble)(tmpvals.preview_width),0.0,1.0);
		toyvals.start[1]=CLAMP((gdouble)active_knot_y/(gdouble)(tmpvals.preview_height),0.0,1.0);
		gtk_adjustment_set_value(knotvals.startx,toyvals.start[0]);
		gtk_adjustment_set_value(knotvals.starty,toyvals.start[1]);
		knotvals.coordinates[0]=active_knot_x;
		knotvals.coordinates[1]=active_knot_y;
		}
	else if (knotvals.inknot2)
		{
		if (both)
			{
			if (knotvals.difference_x>=0)
				{
				if (active_knot_x>=w-1)
					{
					active_knot_x=w-1;
					other_knot_x=w-knotvals.difference_x-1;
					}
				else if (active_knot_x<=knotvals.difference_x)
					{
					active_knot_x=knotvals.difference_x;
					other_knot_x=0;
					}
				}
			else	//knotvals.difference_x<0
				{
				if (active_knot_x>=w+knotvals.difference_x-1) 
					{					
					active_knot_x=w+knotvals.difference_x-1;
					other_knot_x=w-1;
					}
				else if (active_knot_x<=0)
					{
					active_knot_x=0;
					other_knot_x=-knotvals.difference_x;
					}
				}
			if (knotvals.difference_y>=0)
				{
				if (active_knot_y>=h-1) 
					{
					active_knot_y=h-1;
					other_knot_y=h-knotvals.difference_y-1;
					}
				else if (active_knot_y<=knotvals.difference_y)
					{
					active_knot_y=knotvals.difference_y;
					other_knot_y=0;	
					}
				}
			else //knotvals.difference_y<0
				{
				if (active_knot_y>=h+knotvals.difference_y-1) 
					{					
					active_knot_y=h+knotvals.difference_y-1;
					other_knot_y=h-1;
					}
				else if (active_knot_y<=0)
					{
					active_knot_y=0;
					other_knot_y=-knotvals.difference_y;
					}
				}
			toyvals.start[0]=CLAMP((gdouble)other_knot_x/(gdouble)(tmpvals.preview_width),0.0,1.0);
			toyvals.start[1]=CLAMP((gdouble)other_knot_y/(gdouble)(tmpvals.preview_height),0.0,1.0);
			gtk_adjustment_set_value(knotvals.startx,toyvals.start[0]);
			gtk_adjustment_set_value(knotvals.starty,toyvals.start[1]);
			knotvals.coordinates[0]=other_knot_x;
			knotvals.coordinates[1]=other_knot_y;
			}
		toyvals.end[0]=CLAMP((gdouble)active_knot_x/(gdouble)(tmpvals.preview_width),0.0,1.0);
		toyvals.end[1]=CLAMP((gdouble)active_knot_y/(gdouble)(tmpvals.preview_height),0.0,1.0);
		gtk_adjustment_set_value(knotvals.endx,toyvals.end[0]);
		gtk_adjustment_set_value(knotvals.endy,toyvals.end[1]);
		knotvals.coordinates[2]=active_knot_x;
		knotvals.coordinates[3]=active_knot_y;
		}
}

gboolean pressed_keys(GtkWidget *widget,GdkEventKey *key)
{
	GdkCursor *cursor;
	GdkDisplay *display;
	GdkModifierType state;
				
	if (key->keyval==GDK_space)
		{	
		knotvals.gradient_hidden=!knotvals.gradient_hidden;
		display=gtk_widget_get_display(tmpvals.preview);
		gdk_window_get_pointer(tmpvals.preview->window,NULL,NULL,&state);
		if ((knotvals.inknot1||knotvals.inknot2))
			{				
			if (tmpvals.preview&&knotvals.gradient_hidden)
				{
				cursor=gdk_cursor_new_for_display(display,GDK_CROSSHAIR);
				gdk_window_set_cursor(tmpvals.preview->window,cursor);
				gdk_cursor_unref(cursor);
				}
			else if (tmpvals.preview&&!knotvals.gradient_hidden)
				{
				cursor=gdk_cursor_new_for_display(display,(state&GDK_SHIFT_MASK)?GDK_FLEUR:GDK_HAND1);
				gdk_window_set_cursor(tmpvals.preview->window,cursor);
				gdk_cursor_unref(cursor);
				}
			}
		else if ((!knotvals.inknot1||!knotvals.inknot2)&&tmpvals.preview)
			{
			cursor=gdk_cursor_new_for_display(display,GDK_CROSSHAIR);
			gdk_window_set_cursor(tmpvals.preview->window,cursor);
			gdk_cursor_unref(cursor);
			}
		// redraw preview
		preview_idle=1;	
		toy_value_changed(NULL);
		return TRUE;	
		}
	else if 	((key->keyval==GDK_KEY_Shift_L)||(key->keyval==GDK_KEY_Shift_R))
		{
		display=gtk_widget_get_display(tmpvals.preview);
		if ((knotvals.inknot1||knotvals.inknot2)&&tmpvals.preview)
			{
			cursor=gdk_cursor_new_for_display(display,GDK_FLEUR);
			gdk_window_set_cursor(tmpvals.preview->window,cursor);
			gdk_cursor_unref(cursor);
			}
		else if ((!knotvals.inknot1&&!knotvals.inknot2)&&tmpvals.preview)
			{	
			cursor=gdk_cursor_new_for_display(display,GDK_CROSSHAIR);
			gdk_window_set_cursor(tmpvals.preview->window,cursor);
			gdk_cursor_unref(cursor);
			}	
		return TRUE;	
		}		
	return FALSE;
}

gboolean released_keys(GtkWidget *widget,GdkEventKey *key)
{
	if ((key->keyval==GDK_KEY_Shift_L)||(key->keyval==GDK_KEY_Shift_R))
		{
		if ((knotvals.inknot1||knotvals.inknot2)&&tmpvals.preview)
			{	
			GdkDisplay *display=gtk_widget_get_display(tmpvals.preview);
			GdkCursor *cursor=gdk_cursor_new_for_display(display,GDK_HAND1);
			gdk_window_set_cursor(tmpvals.preview->window,cursor);
			gdk_cursor_unref(cursor);
			return TRUE;
			}
		}
	return FALSE;
}
	
void toy_value_changed(GtkWidget *widget)
{
	toy();
	toy_update_preview(0,0,tmpvals.preview_width,tmpvals.preview_height);
	preview_idle=0;
}

void toy_update_preview(gint x0,gint y0,gint x1,gint y1)
{
	if (!preview_idle) {return;}
	gimp_preview_area_draw(GIMP_PREVIEW_AREA(tmpvals.preview),x0,y0,x1-x0,y1-y0,gimp_drawable_type(tmpvals.drawable->drawable_id),tmpvals.preview_buffer,tmpvals.preview_width*gimp_drawable_bpp(tmpvals.drawable->drawable_id));
	draw_gradient(tmpvals.drawable,GIMP_PREVIEW_AREA(tmpvals.preview));
}

void gimp_int_adjustment_update_cs(GtkAdjustment *adjustment,gpointer data)
{
	gint *val=(gint *)data;
	*val=RINT(gtk_adjustment_get_value(adjustment));
	tmpvals.radius_changed=TRUE;
	tmpvals.contrastsaturation_changed=TRUE;
	if (preview_idle) {g_source_remove(preview_idle);}
	preview_idle=g_idle_add_full(G_PRIORITY_LOW,(GSourceFunc)toy_value_changed,NULL,NULL);
}

void gimp_double_adjustment_update_r(GtkAdjustment *adjustment,gpointer data)
{
	gdouble *val=(gdouble*)data;
	*val=gtk_adjustment_get_value(adjustment);
	tmpvals.radius_changed=TRUE;
	if (preview_idle) {g_source_remove(preview_idle);}
	preview_idle=g_idle_add_full(G_PRIORITY_LOW,(GSourceFunc)toy_value_changed,NULL,NULL);
}

void gimp_double_adjustment_update_m(GtkAdjustment *adjustment,gpointer data)
{
	gdouble *val=(gdouble*)data;
	*val=gtk_adjustment_get_value(adjustment);
	if (adjustment==knotvals.startx)
		{
		knotvals.coordinates[0]=(gint)(ROUND(toyvals.start[0]*tmpvals.preview_width));
		knotvals.difference_x=knotvals.coordinates[2]-knotvals.coordinates[0];
		}
	else if (adjustment==knotvals.starty)
		{
		knotvals.coordinates[1]=(gint)(ROUND(toyvals.start[1]*tmpvals.preview_height));
		knotvals.difference_y=knotvals.coordinates[3]-knotvals.coordinates[1];
		}	
	else if (adjustment==knotvals.endx)
		{
		knotvals.coordinates[2]=(gint)(ROUND(toyvals.end[0]*tmpvals.preview_width));
		knotvals.difference_x=knotvals.coordinates[2]-knotvals.coordinates[0];
		}	
	else if (adjustment==knotvals.endy)
		{
		knotvals.coordinates[3]=(gint)(ROUND(toyvals.end[1]*tmpvals.preview_height));
		knotvals.difference_y=knotvals.coordinates[3]-knotvals.coordinates[1];
		}	
	tmpvals.mask_changed=TRUE;
	if (preview_idle) {g_source_remove(preview_idle);}
	preview_idle=g_idle_add_full(G_PRIORITY_LOW,(GSourceFunc)toy_value_changed,NULL,NULL);
}

